/*
 * Copyright 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

/*#include <linux/fs.h>*/
/*#include <linux/malloc.h>*/
#include <linux/sched.h>
#include <linux/reiserfs_fs.h>

#else

#include "nokernel.h"

#endif

/* This is a modification of the lock_buffer code that takes an
   arbitrary lock ulong, bitnumber and wait queue pointer rather than
   being buffer_head specific.  (Undoubtedly one of the SMP guys is
   writting something that will replace this but it was not in the
   code that I have seen.)  Note that lock_buffer has buffer specific
   stuff that this does not do and so this cannot replace that code.
   If someone writes bitops for shorts I'll write this for shorts.
   -Hans */

int __wait_on_ulong_lock(struct super_block *, unsigned short);
extern inline int wait_on_ulong_lock( struct super_block *, unsigned short);
inline int ulong_lock_locked(struct super_block *, unsigned short);
void unlock_ulong_lock(struct super_block *, unsigned short);

/*inline int lock_ulong_lock(unsigned long * lock, unsigned short bitnumber, struct  wait_queue ** l_wait)*/
inline int lock_ulong_lock(struct super_block * s, unsigned short bitnumber)
{
  int repeat = 0;

  if (test_and_set_bit(bitnumber, &(s->u.reiserfs_sb.read_lock)))
    repeat = __wait_on_ulong_lock(s, bitnumber);
  return repeat;
}

extern inline int wait_on_ulong_lock(struct super_block * s, unsigned short bitnumber)
{
  if (ulong_lock_locked(s, bitnumber))
    {
      __wait_on_ulong_lock(s, bitnumber);
      return 1;
    }
  return 0;
}

/* return whether it was able to give you the lock.  Note that it does
   not wait on the lock, and it is used where schedule is a no-no and not
   getting the lock is okay. */
inline int try_ulong_lock(struct super_block * s, unsigned short bitnumber)
{
  return (!test_and_set_bit(bitnumber, &(s->u.reiserfs_sb.read_lock)));
}

int __wait_on_ulong_lock(struct super_block * s, unsigned short bitnumber) 
{
  struct task_struct *tsk = current;
  /*struct wait_queue wait = { current, NULL };*/
  DECLARE_WAITQUEUE(wait, tsk);
  int repeat = 0;

  add_wait_queue(&(s->u.reiserfs_sb.read_wait), &wait);
loop:
  current->state = TASK_UNINTERRUPTIBLE;
  if (!try_ulong_lock(s, bitnumber)) {
    schedule();
    repeat = 1;
    goto loop;
  }
  remove_wait_queue(&(s->u.reiserfs_sb.read_wait), &wait);
  current->state = TASK_RUNNING;
  return repeat;
}

inline int ulong_lock_locked(struct super_block * s, unsigned short bitnumber)
{
  return test_bit(bitnumber, &(s->u.reiserfs_sb.read_lock));
}

/*void unlock_ulong_lock(unsigned long * lock, unsigned short bitnumber, struct  wait_queue ** l_wait)*/
void unlock_ulong_lock(struct super_block * s, unsigned short bitnumber)
{
  if (test_and_clear_bit(bitnumber, &(s->u.reiserfs_sb.read_lock)))  
    wake_up(&(s->u.reiserfs_sb.read_wait));
}
